
from unittest.mock import MagicMock, patch
from urllib.parse import urlparse

from odoo import http
from odoo.tools import mute_logger
from odoo.tests import common, tagged
from odoo.addons.saas_trial.tests.common import TestTrialCommon


@tagged('post_install', '-at_install', 'post_saas_trial_creation')
class TestAuth(TestTrialCommon):

    @classmethod
    def setUpClass(cls):
        super(TestAuth, cls).setUpClass()
        cls.user_customer = cls.env['res.users'].with_context(no_reset_password=True).create({
            'name': "Customer",
            'login': 'customer1@example.net',
            'password': 'customer',
            'tz': 'Europe/Brussels',
            'oauth_uid': "-1",
        })

    def _mock_request_for_trial_login(self):
        login_rejected = False
        mock_trial_login_results = {
            # interactive is True: allow normal password only
            (True, 'customer-odoo-com-passwd'): {'totp_enabled': False},
            (True, 'customer-api-key'): login_rejected,
            (True, 'support-agent/agent-passwd'): {'totp_enabled': False, 'support': True},
            (True, 'support-agent/agent-api-key'): login_rejected,
            (True, 'support-agent-2fa/agent-passwd'): {'totp_enabled': True, 'support': True},
            (True, 'support-agent-2fa/agent-api-key'): login_rejected,

            # interactive is False: allow api-key or normal password (only if no 2fa)
            (False, 'customer-odoo-com-passwd'): {'totp_enabled': False},
            (False, 'customer-api-key'): {'totp_enabled': False},
            (False, 'support-agent/agent-passwd'): {'totp_enabled': False, 'support': True},
            (False, 'support-agent/agent-api-key'): {'totp_enabled': False, 'support': True},
            (False, 'support-agent-2fa/agent-passwd'): login_rejected,
            (False, 'support-agent-2fa/agent-api-key'): {'totp_enabled': True, 'support': True},
        }

        def _request(url, json=None, **kwargs):
            urlp = urlparse(url)
            if urlp.path == "/odoo-enterprise/trial-login":
                assert json["params"]["user_id"] == "-1"
                password = json["params"]['password']
                interactive = json["params"].get('interactive', True)

                trial_login_res = mock_trial_login_results.get((interactive, password), login_rejected)
                if trial_login_res and trial_login_res.get('support'):
                    trial_login_res["impostor_info"] = {'login': password.split('/')[0]}

                mock_result = MagicMock()
                mock_result.json.return_value = {'result': trial_login_res}
                return mock_result
            # for other requests return a dummy (mock) response
            return MagicMock()

        return patch('requests.post', wraps=_request)

    def _web_authenticate(self, db, customer, password):
        """Interactively login using /web/session/authenticate.

        :return: Return the user ID in case of sucessful login, False otherwise
        """
        json_data = {'db': db, 'login': customer, 'password': password}
        resp = self.url_open("/web/session/authenticate", json={'params': json_data})
        resp.raise_for_status()

        try:
            session = http.root.session_store.get(resp.cookies["session_id"])
            return session.uid
        except KeyError:
            return False

    def test_customer_auth(self):
        db = common.get_db_name()
        customer = self.user_customer.login

        with self._mock_request_for_trial_login() as mock_request:
            # # customer: local auth
            uid = self._web_authenticate(db, customer, 'customer')
            self.assertEqual(uid, self.user_customer.id)
            self.assertEqual(mock_request.call_count, 0)

            # customer: allow connection using odoo.com password
            # (password will get sync. to local user)
            mock_request.reset_mock()
            uid = self._web_authenticate(db, customer, 'customer-odoo-com-passwd')
            self.assertEqual(uid, self.user_customer.id)
            self.assertEqual(mock_request.call_count, 1)
            self.assertIn("/odoo-enterprise/trial-login", mock_request.call_args_list[0][0][0])

    def test_customer_auth_rpc(self):
        db = self.env.cr.dbname
        customer = self.user_customer.login
        customer_uid = self.user_customer.id

        with self._mock_request_for_trial_login() as mock_request:

            # customer: local auth
            uid = self.xmlrpc_common.login(db, customer, "customer")
            self.assertEqual(uid, self.user_customer.id)
            self.assertEqual(mock_request.call_count, 0)

            # customer: allow connection using its odoo.com password
            # (password should not be sync. to local user for RPC)
            mock_request.reset_mock()
            uid = self.xmlrpc_common.login(db, customer, "customer-odoo-com-passwd")
            self.assertEqual(uid, self.user_customer.id)
            self.assertEqual(mock_request.call_count, 1)

            # customer: 2nd connection using its odoo.com password
            # (password was not synced, so we should call trial-login again)
            mock_request.reset_mock()
            uid = self.xmlrpc_common.login(db, customer, "customer-odoo-com-passwd")
            self.assertEqual(uid, self.user_customer.id)
            self.assertEqual(mock_request.call_count, 1)

            # customer: allow connection using its odoo.com api-key
            mock_request.reset_mock()
            uid = self.xmlrpc_common.login(db, customer, "customer-api-key")
            self.assertEqual(uid, self.user_customer.id)
            self.assertEqual(mock_request.call_count, 1)

            # customer: 2nd connection using its odoo.com api-key
            # (password was no synced, so we should still call trail-login)
            mock_request.reset_mock()
            uid = self.xmlrpc_common.login(db, customer, "customer-api-key")
            self.assertEqual(uid, self.user_customer.id)
            self.assertEqual(mock_request.call_count, 1)

            # customer: ensure valid credentials from odoo.com are cached
            # so we do not need to call trial login for each RPC
            mock_request.reset_mock()
            for _dummy in range(2):
                uids = self.xmlrpc_object.execute(
                    db, customer_uid, "customer-api-key",
                    "res.users", "search", [('id', '=', customer_uid)]
                )
                self.assertEqual(uids, self.user_customer.ids)
            self.assertEqual(mock_request.call_count, 1)

    @mute_logger("odoo.addons.saas_trial.models.res_users", "odoo.http")
    def test_support_auth(self):
        """Test interactive authentication for support agent"""
        db = common.get_db_name()
        customer = self.user_customer.login

        with self._mock_request_for_trial_login() as mock_request:
            # support: should never be allowed to log-in interactively
            mock_request.reset_mock()
            uid = self._web_authenticate(db, customer, "support-agent-2fa/agent-api-key")
            self.assertFalse(uid)
            self.assertEqual(mock_request.call_count, 1)

            mock_request.reset_mock()
            uid = self._web_authenticate(db, customer, "support-agent-2fa/agent-passwd")
            self.assertFalse(uid)
            self.assertEqual(mock_request.call_count, 1)

            mock_request.reset_mock()
            uid = self._web_authenticate(db, customer, "support-agent/agent-passwd")
            self.assertFalse(uid)
            self.assertEqual(mock_request.call_count, 1)

            mock_request.reset_mock()
            uid = self._web_authenticate(db, customer, "support-agent/agent-api-key")
            self.assertFalse(uid)
            self.assertEqual(mock_request.call_count, 1)

    @mute_logger("odoo.addons.saas_trial.models.res_users")
    def test_support_auth_rpc(self):
        """Test non-interactive authentication for support agent"""
        db = common.get_db_name()
        customer = self.user_customer.login

        with self._mock_request_for_trial_login() as mock_request:
            # support: only agent with 2fa and using api-key should be able to connect
            mock_request.reset_mock()
            uid = self.xmlrpc_common.login(db, customer, "support-agent-2fa/agent-api-key")
            self.assertEqual(uid, self.user_customer.id)
            self.assertEqual(mock_request.call_count, 1)

            mock_request.reset_mock()
            uid = self.xmlrpc_common.login(db, customer, "support-agent-2fa/agent-passwd")
            self.assertEqual(uid, False)
            self.assertEqual(mock_request.call_count, 1)

            mock_request.reset_mock()
            uid = self.xmlrpc_common.login(db, customer, "support-agent/agent-passwd")
            self.assertEqual(uid, False)
            self.assertEqual(mock_request.call_count, 1)

            mock_request.reset_mock()
            uid = self.xmlrpc_common.login(db, customer, "support-agent/api-key")
            self.assertEqual(uid, False)
            self.assertEqual(mock_request.call_count, 1)
